#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <algorithm>
#include <math.h>
#include <iomanip>
#include <unordered_map>
#include <set>
#include <tuple>
#include <sstream>
#include <random>

#include "Parameters.h"
#include "Pedigree.h"

using namespace std;
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////      Class Functions       ////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
parameters::parameters(){}
parameters::~parameters(){}
void parameters::Updatedatafile(std::string temp){datafile = temp;}
void parameters::Updatepedigreefile(std::string temp){pedigreefile = temp;}
void parameters::Updatesamples(int temp){samples = temp;}
void parameters::Updateburnin(int temp){burnin = temp;}
void parameters::Updatethin(int temp){thin = temp;}
void parameters::Updateouteffectfile(std::string temp){outeffectfile = temp;}
void parameters::Updateoutvariancefile(std::string temp){outvariancefile = temp;}
void parameters::Updatethreads(int temp){threads = temp;}

/*******************************************************************/
/*******************************************************************/
/** Read in Parameter File and declare parameters and define model */
/*******************************************************************/
/*******************************************************************/
void read_generate_parameters(parameters &Parameters, string parameterfile, ostream& logfile)
{
    /* read parameters file */
    vector <string> parm;
    string parline;
    ifstream parfile;
    parfile.open(parameterfile);
    if(parfile.fail()){cout << "Parameter file not found. Check log file." << endl; exit (EXIT_FAILURE);}
    while (getline(parfile,parline)){parm.push_back(parline);} /* Stores in vector and each new line push back to next space */
    logfile << "===============================================" << endl;
    logfile << "==        Read in Parameters from file       ==" << endl;
    logfile << "===============================================" << endl;
    logfile << "Name of parameter file was: '"+parameterfile+"'" << endl;
    logfile << " - Parameters Involving the Model: " << endl;
    int search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("DATAFILE:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updatedatafile(parm[search]); parm.erase(parm.begin()+search);
            logfile << "    - Name of Data File: '"<<Parameters.getdatafile()<<"'."<< endl; break;
        }
        search++; if(search >= parm.size()){cout << endl << "Couldn't find 'DATAFILE:' variable in parameter file!" << endl; exit (EXIT_FAILURE);}
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("TRAIT(S)_COLUMN:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(":", 0); parm[search].erase(0, pos + 1);
            vector < string > solvervariables(10,"");
            for(int i = 0; i < 10; i++)
            {
                size_t pos = parm[search].find(" ",0);
                solvervariables[i] = parm[search].substr(0,pos);
                solvervariables[i].erase(remove(solvervariables[i].begin(), solvervariables[i].end(), ' '),solvervariables[i].end());
                if(pos != std::string::npos){parm[search].erase(0, pos + 1);}
                if(pos == std::string::npos){parm[search].clear(); i = 10;}
            }
            int start = 0;
            while(start < solvervariables.size())
            {
                if(solvervariables[start] == ""){solvervariables.erase(solvervariables.begin()+start);}
                if(solvervariables[start] != ""){start++;}
            }
            for(int i = 0; i < solvervariables.size(); i++){Parameters.add_trait_col(atoi(solvervariables[i].c_str()));}
            logfile << "    - Column(s) where phenotypes are located:";
            for(int i = 0; i < Parameters.get_trait_col().size(); i++){logfile << " " << (Parameters.get_trait_col())[i];}
            logfile << "." << endl;
            break;
        }
        search++; if(search >= parm.size()){cout << endl << "Couldn't find 'TRAIT(S)_COLUMN:' variable in parameter file!" << endl; exit (EXIT_FAILURE);}
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("RESIDUAL_VARIANCE:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(":", 0); parm[search].erase(0, pos + 1);
            vector < string > solvervariables(10,"");
            for(int i = 0; i < 10; i++)
            {
                size_t pos = parm[search].find(" ",0);
                solvervariables[i] = parm[search].substr(0,pos);
                solvervariables[i].erase(remove(solvervariables[i].begin(), solvervariables[i].end(), ' '),solvervariables[i].end());
                if(pos != std::string::npos){parm[search].erase(0, pos + 1);}
                if(pos == std::string::npos){parm[search].clear(); i = 10;}
            }
            int start = 0;
            while(start < solvervariables.size())
            {
                if(solvervariables[start] == ""){solvervariables.erase(solvervariables.begin()+start);}
                if(solvervariables[start] != ""){start++;}
            }
            /* Check to determine if number of parameters is equal to number of phenotypes */
            if((Parameters.get_trait_col()).size() == 1 && solvervariables.size() != 1)
            {
                cout << endl <<"Incorrect number of residual variance starting values based on the number of phenotypes!" << endl; exit (EXIT_FAILURE);
            }
            for(int i = 0; i < solvervariables.size(); i++){Parameters.add_residual_start(atof(solvervariables[i].c_str()));}
            logfile << "    - Residual Variance Structure and Starting Parameters:" << endl;
            if((Parameters.get_trait_col()).size() == 1){logfile << "             - " << (Parameters.get_rresidual_start())[0] << endl;}
            break;
        }
        search++; if(search >= parm.size()){cout << endl << "Couldn't find 'RESIDUAL_VARIANCE:' variable in parameter file!" << endl; exit (EXIT_FAILURE);}
    }
    /**********************************/
    /** Search For All Fixed Effects **/
    /**********************************/
    string stopsearching = "NO";
    logfile << "    - Fixed Effects Columns and Classifications:" << endl;
    while(stopsearching == "NO")
    {
        search = 0;
        while(1)
        {
            size_t fnd = parm[search].find("EFFECT_Fixed:");
            if(fnd!=std::string::npos)
            {
                size_t pos = parm[search].find(":", 0); parm[search].erase(0, pos + 1);
                vector < string > solvervariables(10,"");
                for(int i = 0; i < 10; i++)
                {
                    size_t pos = parm[search].find(" ",0);
                    solvervariables[i] = parm[search].substr(0,pos);
                    solvervariables[i].erase(remove(solvervariables[i].begin(), solvervariables[i].end(), ' '),solvervariables[i].end());
                    if(pos != std::string::npos){parm[search].erase(0, pos + 1);}
                    if(pos == std::string::npos){parm[search].clear(); i = 10;}
                }
                int start = 0;
                while(start < solvervariables.size())
                {
                    if(solvervariables[start] == ""){solvervariables.erase(solvervariables.begin()+start);}
                    if(solvervariables[start] != ""){start++;}
                }
                /* Check to determine if number of parameters is equal to number of phenotypes */
                if(solvervariables.size() != 2){cout << endl <<"Incorrect number of parameters in 'EFFECT_Fixed'!" << endl; exit (EXIT_FAILURE);}
                Parameters.add_fixed_col(atoi(solvervariables[0].c_str()));
                Parameters.add_fixed_type(solvervariables[1]);
                logfile << "        - Fixed Effect " << (Parameters.get_fixed_col()).size() << ": Column ";
                logfile << (Parameters.get_fixed_col())[(Parameters.get_fixed_col()).size()-1] << " classified as '";
                logfile << (Parameters.get_fixed_type())[(Parameters.get_fixed_type()).size()-1]  << "'." << endl;
                break;
            }
            search++;
            if(search >= parm.size()){stopsearching = "YES"; break;}
        }
    }
    /***********************************/
    /** Search For All Random Effects **/
    /***********************************/
    stopsearching = "NO";
    logfile << "    - Random Effects Columns and Classifications:" << endl;
    while(stopsearching == "NO")
    {
        search = 0;
        while(1)
        {
            size_t fnd = parm[search].find("EFFECT_Random:");
            if(fnd!=std::string::npos)
            {
                size_t pos = parm[search].find(":", 0); parm[search].erase(0, pos + 1);
                vector < string > solvervariables(10,"");
                for(int i = 0; i < 10; i++)
                {
                    size_t pos = parm[search].find(" ",0);
                    solvervariables[i] = parm[search].substr(0,pos);
                    solvervariables[i].erase(remove(solvervariables[i].begin(), solvervariables[i].end(), ' '),solvervariables[i].end());
                    if(pos != std::string::npos){parm[search].erase(0, pos + 1);}
                    if(pos == std::string::npos){parm[search].clear(); i = 10;}
                }
                int start = 0;
                while(start < solvervariables.size())
                {
                    if(solvervariables[start] == ""){solvervariables.erase(solvervariables.begin()+start);}
                    if(solvervariables[start] != ""){start++;}
                }
                /* Check to determine if number of parameters is equal to number of phenotypes */
                if(solvervariables.size() != 3){cout << endl <<"Incorrect number of parameters in 'EFFECT_Random'!" << endl; exit (EXIT_FAILURE);}
                Parameters.add_random_col(atoi(solvervariables[0].c_str()));
                Parameters.add_random_type(solvervariables[1]);
                Parameters.add_random_start(atof(solvervariables[2].c_str()));
                logfile << "        - Random Effect " << (Parameters.get_random_col()).size() << ": Column ";
                logfile << (Parameters.get_random_col())[(Parameters.get_random_col()).size()-1] << ", covariance structure is a '";
                logfile << (Parameters.get_random_type())[(Parameters.get_random_type()).size()-1]  << "' and the starting value is '";
                logfile << (Parameters.get_random_start())[(Parameters.get_random_start()).size()-1]  << "'." << endl;
                break;
            }
            search++;
            if(search >= parm.size()){stopsearching = "YES"; break;}
        }
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("PEDIGREE:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updatepedigreefile(parm[search]); parm.erase(parm.begin()+search);
            logfile << "    - Name of Pedigree File: '"<<Parameters.getpedigreefile()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size()){cout << endl << "Couldn't find 'PEDIGREE:' variable in parameter file!" << endl; exit (EXIT_FAILURE);}
    }
    logfile << " - Parameters Involving Gibbs Sampling: " << endl;
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("THREADS:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updatethreads(atoi(parm[search].c_str())); parm.erase(parm.begin()+search);
            logfile << "    - Number of threads used in program: '"<<Parameters.getthreads()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size())
        {
            Parameters.Updatethreads(1);
            logfile << "    - Number of threads used in program: '1' (Default)."<< endl; break;
        }
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("SAMPLES:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updatesamples(atoi(parm[search].c_str())); parm.erase(parm.begin()+search);
            logfile << "    - Number of Gibbs Samples: '"<<Parameters.getsamples()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size()){cout << endl << "Couldn't find 'SAMPLES:' variable in parameter file!" << endl; exit (EXIT_FAILURE);}
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("BURN_IN:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updateburnin(atoi(parm[search].c_str())); parm.erase(parm.begin()+search);
            logfile << "    - Number of Samples to Burn In: '"<<Parameters.getburnin()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size()){cout << endl << "Couldn't find 'BURN_IN:' variable in parameter file!" << endl; exit (EXIT_FAILURE);}
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("THIN:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updatethin(atoi(parm[search].c_str())); parm.erase(parm.begin()+search);
            logfile << "    - Thinning Rate: '"<<Parameters.getthin()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size()){Parameters.Updatethin(1); logfile << "    - Thinning Rate: '1' (Default)."<< endl; break;}
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("OUTEFFECT:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updateouteffectfile(parm[search].c_str()); parm.erase(parm.begin()+search);
            logfile << "    - Solutions stored in: '"<<Parameters.getouteffectfile()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size())
        {
            Parameters.Updateouteffectfile("solutions_effects");
            logfile << "    - Solutions stored in: 'solutions_effects' (Default)."<< endl; break;
        }
    }
    search = 0;
    while(1)
    {
        size_t fnd = parm[search].find("OUTVARIANCE:");
        if(fnd!=std::string::npos)
        {
            size_t pos = parm[search].find(": ", 0); parm[search].erase(0, pos + 1);
            parm[search].erase(remove(parm[search].begin(), parm[search].end(), ' '), parm[search].end());
            Parameters.Updateoutvariancefile(parm[search].c_str()); parm.erase(parm.begin()+search);
            logfile << "    - Variance components stored in: '"<<Parameters.getoutvariancefile()<<"'."<< endl; break;
        }
        search++;
        if(search >= parm.size())
        {
            Parameters.Updateoutvariancefile("solutions_variances");
            logfile << "    - Variance components stored in: 'solutions_variances' (Default)."<< endl; break;
        }
    }
    logfile << "All parameters read in." << endl << endl;
}
