#include <iostream>
#include <vector>
#include <string>
#ifndef ParameterClass_H_
#define ParameterClass_H_

using namespace std;

/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
// Class for parameters used in the gibbs sampling program //
/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
class parameters
{
private:
    int threads = 1;                                            /* Number of Cores in program */
    string datafile = "";                                       /* Datafile */
    vector <int> trait_col = vector<int>(0);                    /* Columns where trait appears */
    vector <int> fixed_col = vector<int>(0);                    /* Columns where fixed effects appear */
    vector <string> fixed_type = vector <string>(0);            /* Type of fixed effect int, class covariate */
    vector <int> random_col = vector<int>(0);                   /* Columns where random effects appear */
    vector <string> random_type = vector <string>(0);           /* Type of random effect Ainv, Identity */
    vector <double> random_start = vector <double>(0);          /* Starting Value for random effect variance */
    string pedigreefile = "";                                   /* Pedigree file */
    vector <double> residual_start = vector <double>(0);        /* Starting Value for residual variance */
    int samples = 0;                                            /* Number of Samples */
    int burnin = 0;                                             /* Number of Samples to Burn-In */
    int thin = 0;                                               /* Keep every nth sample to estimate parameters */
    string outeffectfile = "solutions_effects";                  /* Save mean effects to a file */
    string outvariancefile = "solutions_variances";              /* Save variance to a file */
public:
    parameters();
    ~parameters();
    int getthreads(){return threads;}
    void Updatethreads(int temp);
    /* Data File Name */
    std::string getdatafile(){return datafile;}
    void Updatedatafile(std::string temp);
    /* Pedigree File Name */
    std::string getpedigreefile(){return pedigreefile;}
    void Updatepedigreefile(std::string temp);
    /* Columns where trait appears */
    const vector <int>& get_trait_col() const{return trait_col;}
    void add_trait_col(int x){trait_col.push_back(x);}
    /* Columns where fixed effects appear */
    const vector <int>& get_fixed_col() const{return fixed_col;}
    void add_fixed_col(int x){fixed_col.push_back(x);}
    /* Type of Fixed effect */
    const vector <string>& get_fixed_type() const{return fixed_type;}
    void add_fixed_type(string x){fixed_type.push_back(x);}
    /* Columns where random effects appear */
    const vector <int>& get_random_col() const{return random_col;}
    void add_random_col(int x){random_col.push_back(x);}
    /* Type of Random effect */
    const vector <string>& get_random_type() const{return random_type;}
    void add_random_type(string x){random_type.push_back(x);}
    /* Random Start Values */
    const vector <double>& get_random_start() const{return random_start;}
    void add_random_start(double x){random_start.push_back(x);}
    /* Starting Value for residual variance */
    const vector <double>& get_rresidual_start() const{return residual_start;}
    void add_residual_start(double x){residual_start.push_back(x);}
    /* Samples, Burn-in and Thin Rate */
    int getsamples(){return samples;}
    void Updatesamples(int temp);
    int getburnin(){return burnin;}
    void Updateburnin(int temp);
    int getthin(){return thin;}
    void Updatethin(int temp);
    /* Save Output to Files */
    std::string getouteffectfile(){return outeffectfile;}
    void Updateouteffectfile(std::string temp);
    std::string getoutvariancefile(){return outvariancefile;}
    void Updateoutvariancefile(std::string temp);
};

#endif
