#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <algorithm>
#include <math.h>
#include <iomanip>
#include <unordered_map>
#include <set>
#include <tuple>
#include <sstream>
#include <random>

#include "Parameters.h"
#include "Pedigree.h"

using namespace std;
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////      Class Functions       ////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
Pedigree_Class::Pedigree_Class(){animal = ""; sire = ""; dam = ""; renum_anim = 0; generation = 0;}
Pedigree_Class::Pedigree_Class(string anim, string si, string da, int reanim, int gen){animal = anim; sire = si; dam = da; renum_anim = reanim; generation = gen;}
Pedigree_Class::~Pedigree_Class(){};
// Functions
void Pedigree_Class::Update_Generation(int temp){generation = temp;}
void Pedigree_Class::Update_Renum_Anim(int temp){renum_anim = temp;}
bool sortByGen(const Pedigree_Class &lhs, const Pedigree_Class &rhs) {return lhs.generation > rhs.generation;} // Sort objects

/**********************************************/
/**********************************************/
/** Read in pedigree, sort and create inverse */
/**********************************************/
/**********************************************/
void read_sort_ainv(vector < Pedigree_Class > &pedigree, vector < tuple <int,int,double> > &sprelationshipinv, parameters &Parameters,vector<int> &nonzerorows)
{
    unordered_map <string,int> linker;                  /* Hash-table data structure for faster lookup */
std:unordered_map<string,int>::const_iterator got;  /* Search if key exists */
    string line,tempanimal,tempsire,tempdam;            /* Initialize temp variables */
    ifstream infile;
    infile.open((Parameters.getpedigreefile()).c_str());
    if(infile.fail()){cout << "Pedigree File Doesn't Exist!\n"; exit (EXIT_FAILURE);}
    while (getline(infile,line))
    {
        size_t pos = line.find(" ", 0); tempanimal = line.substr(0,pos); line.erase(0, pos+1);
        /* If have blank lines at the end of file don't add them */
        if(tempanimal != "")
        {
            pos = line.find(" ", 0); tempsire = line.substr(0,pos); line.erase(0, pos+1); tempdam = line;
            /* Weird stuff happens at the end of the row if reading in windows file versus mac file*/
            tempdam.erase(tempdam.find_last_not_of(" \n\r\t")+1);
            tempdam.erase(remove(tempdam.begin(),tempdam.end(),' '),tempdam.end());
            /* if you find one of the possible missing value change it to 0 */
            if(tempsire == "." || tempsire == "-"){tempsire = "0";}
            if(tempdam == "." || tempdam == "-"){tempdam = "0";}
            /* check to make sure animal doesn't already exist in pedigree; if so exit out of program */
            got = linker.find(tempanimal);
            if(got == linker.end())     /* If not found then add */
            {
                /* save in class */
                Pedigree_Class pedigreetemp(tempanimal,tempsire,tempdam,1,1);
                pedigree.push_back(pedigreetemp);
                linker.insert({tempanimal,pedigree.size()-1});
            } else {cout << "Exit Program. Animal (" << tempanimal << ") appears more than once in pedigree file!!" << endl; exit (EXIT_FAILURE);}
        }
    }
    /* check to make sure all sire and dams are represented as an animal and if missing add (assume both parents are unknown) */
    for(int i = 0; i < pedigree.size(); i++)
    {
        if(pedigree[i].getsire() != "0" || pedigree[i].getdam() != "0")
        {
            if(pedigree[i].getsire() != "0")
            {
                /* See if it sire exists as an animal */
                got = linker.find(pedigree[i].getsire());
                if(got == linker.end())     /* If not found then add */
                {
                    Pedigree_Class pedigreetemp(pedigree[i].getsire(),"0","0",1,1);
                    pedigree.push_back(pedigreetemp);
                    linker.insert({pedigree[i].getsire(),pedigree.size()-1});
                }
            }
            if(pedigree[i].getdam() != "0")
            {
                /* See if it dam exists as an animal */
                got = linker.find(pedigree[i].getdam());
                if(got == linker.end())     /* If not found then add */
                {
                    Pedigree_Class pedigreetemp(pedigree[i].getdam(),"0","0",1,1);
                    pedigree.push_back(pedigreetemp);
                    linker.insert({pedigree[i].getdam(),(pedigree.size()-1)});
                }
            }
        }
    }
    /* check to make an animal doesn't have a value of 0 */
    for(int i = 0; i < pedigree.size(); i++)
    {
        if(pedigree[i].getanim() == "0"){cout << endl << "Exit Program. Animal ID can't be '0', line " << i+1 << endl; exit (EXIT_FAILURE);}
        if(pedigree[i].getanim() == ""){cout << endl << "Exit Program. Animal ID can't be '', line " << i+1 << endl; exit (EXIT_FAILURE);}
    }
    /* check to make sure an animal only is a sire or a dam */
    vector < int > sirecount(pedigree.size(),0);    /* Number of times animal appears as a sire */
    vector < int > damcount(pedigree.size(),0);     /* Number of time animal appears as a dam */
    /* loop through and add a count to a given animal and use linker hash table to look up where sire and dam is at */
    for(int i = 0; i < pedigree.size(); i++)
    {
        if(pedigree[i].getsire() != "0" || pedigree[i].getdam() != "0")
        {
            if(pedigree[i].getsire() != "0"){sirecount[linker[pedigree[i].getsire()]] += 1;}
            if(pedigree[i].getdam() != "0"){damcount[linker[pedigree[i].getdam()]] += 1;}
        }
    }
    for(int i = 0; i < sirecount.size(); i++)
    {
        if(sirecount[i] > 0 && damcount[i] > 0){cout << "Exit Program. Animal (" << pedigree[i].getanim() << ") appears as an sire and dam!!" << endl; exit (EXIT_FAILURE);}
    }
    /* Algorithm works by proceeding through the pedigree one animal at a time and all animals begin with generation number = 1. Take current generation and increase it by */
    /* 1 and compare m to the generation number of sire and dam. If it is less than sire and dam then change DD to generation 2; This was taken from Dr. Schaeffer's Notes. */
    int orderiteration = 0; string done = "NO";
    while(done == "NO")
    {
        orderiteration++; int numberchanged = 0;
        for(int i = 0; i < pedigree.size(); i++)
        {
            if(pedigree[i].getsire() != "0" | pedigree[i].getdam() != "0")
            {
                if(pedigree[i].getsire() != "0")
                {
                    if(pedigree[linker[pedigree[i].getsire()]].getgener() <= pedigree[i].getgener())
                    {
                        pedigree[linker[pedigree[i].getsire()]].Update_Generation(pedigree[i].getgener() + 1); numberchanged++;
                    }
                }
                if(pedigree[i].getdam() != "0")
                {
                    if(pedigree[linker[pedigree[i].getdam()]].getgener() <= pedigree[i].getgener())
                    {
                        pedigree[linker[pedigree[i].getdam()]].Update_Generation(pedigree[i].getgener() + 1); numberchanged++;
                    }
                }
            }
        }
        //if(orderiteration % 1 == 0){cout << "   - " << orderiteration << " (" << numberchanged << ")" << endl;}
        if(numberchanged == 0){done = "YES";}
    }
    //sort(pedigree.begin(), pedigree.end(), sortByGen);
    /* Now set renumber ID so they go from 1:number of animals and update linker file since they were sorted*/
    linker.clear(); /* First clear old linker file */
    for(int i = 0; i < pedigree.size(); i++)
    {
        pedigree[i].Update_Renum_Anim(i+1);         /* sets ids to go from 1:n */
        linker.insert({pedigree[i].getanim(),i});   /* updates linker file */
    }
    /* now generate animal, sire and dam vectors to use to generate A based on renum IDs */
    vector <int> renum_animal(pedigree.size(),0); vector <int> renum_sire(pedigree.size(),0); vector <int> renum_dam(pedigree.size(),0);
    /* Now set renumber ID */
    for(int i = 0; i < pedigree.size(); i++)
    {
        renum_animal[i] = pedigree[i].getrenumanim();
        if(pedigree[i].getsire() != "0" | pedigree[i].getdam() != "0")
        {
            if(pedigree[i].getsire() != "0"){renum_sire[i] = pedigree[linker[pedigree[i].getsire()]].getrenumanim();}
            if(pedigree[i].getdam() != "0"){renum_dam[i] = pedigree[linker[pedigree[i].getdam()]].getrenumanim();}
        }
    }
    /* Save lower triangle value as a tuple (int, int, double) and then once finished add them all up and output */
    int animnumb = renum_animal.size();
    vector < double > F((animnumb+1),0.0); vector < double > D(animnumb,0.0);
    /* This it makes so D calculate is correct */
    F[0] = -1;
    for(int k = renum_animal[0]; k < (animnumb+1); k++)                  /* iterate through each row of l */
    {
        vector < double > L(animnumb,0.0);
        vector < double > AN;                                       /* Holds all ancestors of individuals */
        double ai = 0.0;                                            /* ai  is the inbreeding */
        AN.push_back(k);                                            /* push back animal in ancestor */
        L[k-1] = 1.0;
        /* Calculate D */
        D[k-1] = 0.5 - (0.25 * (F[renum_sire[k-1]] + F[renum_dam[k-1]]));
        int j = k;                                          /* start off at K then go down */
        while(AN.size() > 0)
        {
            if((renum_sire[j-1]) != 0){AN.push_back(renum_sire[j-1]); L[renum_sire[j-1]-1] += 0.5 * L[j-1];}        /* if sire is know add to AN and add to L[j] */
            if((renum_dam[j-1]) != 0){AN.push_back(renum_dam[j-1]); L[renum_dam[j-1]-1] += 0.5 * L[j-1];}           /* if dam is known add to AN and add to L[j] */
            ai += (L[j-1] * L[j-1] * D[j-1]);                                                                       /* add to inbreeding value */
            /* Delete j from n */
            int found = 0;
            while(1)
            {
                if(AN[found] == j){AN.erase(AN.begin()+found); break;}
                if(AN[found] != j){found++;}
            }
            /* Find youngest animal in AN to see if it has ancestors*/
            j = -1;
            for(int i = 0; i < AN.size(); i++){if(AN[i] > j){j = AN[i];}}
            /* Erase Duplicates */
            sort(AN.begin(),AN.end());
            AN.erase(unique(AN.begin(),AN.end()),AN.end());
        }
        /* calculate inbreeding value */
        F[k] = ai - 1;
        double bi = (1/sqrt(D[k-1])) * (1/sqrt(D[k-1]));
        if(renum_sire[k-1] != 0 && renum_dam[k-1] != 0)     /* both parent known */
        {
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_animal[k-1]-1),bi));
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_sire[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_sire[k-1]-1),(renum_animal[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_dam[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_dam[k-1]-1),(renum_animal[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_sire[k-1]-1),(renum_sire[k-1]-1),(bi/double(4))));
            sprelationshipinv.emplace_back(std::make_tuple((renum_dam[k-1]-1),(renum_sire[k-1]-1),(bi/double(4))));
            sprelationshipinv.emplace_back(std::make_tuple((renum_sire[k-1]-1),(renum_dam[k-1]-1),(bi/double(4))));
            sprelationshipinv.emplace_back(std::make_tuple((renum_dam[k-1]-1),(renum_dam[k-1]-1),(bi/double(4))));
        }
        if(renum_sire[k-1] != 0 && renum_dam[k-1] == 0)     /* renum_sire parent known */
        {
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_animal[k-1]-1),bi));
            sprelationshipinv.emplace_back(std::make_tuple((renum_sire[k-1]-1),(renum_animal[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_sire[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_sire[k-1]-1),(renum_sire[k-1]-1),(bi/double(4))));
        }
        if(renum_sire[k-1] == 0 && renum_dam[k-1] != 0)     /* renum_dam parent known */
        {
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_animal[k-1]-1),bi));
            sprelationshipinv.emplace_back(std::make_tuple((renum_dam[k-1]-1),(renum_animal[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_dam[k-1]-1),((bi/2)*-1)));
            sprelationshipinv.emplace_back(std::make_tuple((renum_dam[k-1]-1),(renum_dam[k-1]-1),(bi/double(4))));
        }
        if(renum_sire[k-1] == 0 && renum_dam[k-1] == 0) /* both parent unknown */
        {
            sprelationshipinv.emplace_back(std::make_tuple((renum_animal[k-1]-1),(renum_animal[k-1]-1),bi));
        }
    }
    for(int i = 0; i < pedigree.size(); i++){nonzerorows.push_back(0);}
    for (int i = 0; i < sprelationshipinv.size(); i++){nonzerorows[get<1>(sprelationshipinv[i])] += 1;}
    //for(int i = 0; i < nonzerorows.size(); i++){cout << nonzerorows[i] << " ";}
    //cout << endl;
}


